import io.micrometer.context.ContextRegistry;
import io.micrometer.observation.Observation;
import io.micrometer.observation.ObservationRegistry;
import io.micrometer.tracing.BaggageInScope;
import io.micrometer.tracing.Tracer;
import io.micrometer.tracing.contextpropagation.ObservationAwareSpanThreadLocalAccessor;
import jakarta.annotation.PostConstruct;
import lombok.extern.log4j.Log4j2;
import org.junit.jupiter.api.MethodOrderer;
import org.junit.jupiter.api.Order;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestMethodOrder;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.EnableAutoConfiguration;
import org.springframework.boot.test.autoconfigure.actuate.observability.AutoConfigureObservability;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.annotation.Configuration;
import org.springframework.test.context.TestPropertySource;

import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import static java.util.Optional.ofNullable;
import static org.assertj.core.api.Assertions.assertThat;

@SpringBootTest(classes = TestExample.TestConfig.class)
@AutoConfigureObservability
@TestPropertySource(properties =
        "logging.pattern.level=[${spring.application.name:},%X{traceId:-},%X{spanId:-}]"
)
@TestMethodOrder(MethodOrderer.OrderAnnotation.class)
@Log4j2
class TestExample {

    @Configuration
    @EnableAutoConfiguration
    static class TestConfig{
        @Autowired
        Tracer tracer;

        @PostConstruct
        public void registerObservationAwareThreadLocalAccessor() {
                ContextRegistry.getInstance()
                        .registerThreadLocalAccessor(new ObservationAwareSpanThreadLocalAccessor(tracer));
        }
    }

    @Autowired
    Tracer tracer;

    @Autowired
    ObservationRegistry observationRegistry;

    @Test
    @Order(1)
    void this_test_behaves_as_expected(){
        assertThat(tracer.currentTraceContext().context()).isNull();
        assertThat(tracer.getAllBaggage()).isEmpty();
        logTracingStatus();

        Observation.createNotStarted("this_test_behaves_as_expected", observationRegistry)
                .observe(() -> {
                    assertThat(tracer.currentTraceContext().context()).isNotNull();
                    logTracingStatus();

                    BaggageInScope baggageInScope = tracer.createBaggageInScope("hello", "world");
                    assertThat(tracer.getAllBaggage()).containsOnlyKeys("hello");
                    baggageInScope.close();
                });

        assertThat(tracer.currentTraceContext().context()).isNull();
        logTracingStatus();
        assertThat(tracer.getAllBaggage()).isEmpty();
    }

    static Stream<Arguments> this_test_does_not(){
        return Stream.of(
                Arguments.of(Map.of("hello", "world", "foo", "bar"), 1),
                Arguments.of(Map.of("foo", "bar"), 2),
                Arguments.of(Map.of(), 3)
                );
    }

    @ParameterizedTest
    @MethodSource
    @Order(2)
    void this_test_does_not(Map<String, String> baggage, int id){
//        assertThat(tracer.currentTraceContext().context()).isNull();
//        assertThat(tracer.getAllBaggage()).isEmpty();
        logTracingStatus();

        Observation.createNotStarted("this_test_does_not-" + id, observationRegistry)
                .observe(() -> {
                    assertThat(tracer.currentTraceContext().context()).isNotNull();
                    logTracingStatus();

                    List<BaggageInScope> baggageInScope = baggage.entrySet().stream().map(entry -> tracer.createBaggageInScope(entry.getKey(), entry.getValue())).toList();
                    assertThat(tracer.getAllBaggage()).containsExactlyInAnyOrderEntriesOf(baggage);
                    baggageInScope.forEach(BaggageInScope::close);
                });

        assertThat(tracer.currentTraceContext().context()).isNull();
        logTracingStatus();
        assertThat(tracer.getAllBaggage()).isEmpty();
    }

    @Test
    @Order(3)
    void following_tests_fail_because_trace_is_not_null_at_start(){
        this_test_behaves_as_expected();
    }

    private void logTracingStatus(){
        ofNullable(tracer.currentTraceContext().context())
                .ifPresentOrElse(
                        context -> log.info("<------- tracing context present"),
                        () -> log.info("no tracing context present")
                );
    }
}
